# The Workflow Management System *WMS-light*

## 1 OVERVIEW

Many scientific application are build by interconnecting several functionally-independent components within a common data- and workflow scenario. Such applications have to deal with the issues of the coordinated execution and organization of the data exchange between those components.


| *A workflow example:*              | |
|------------------------------------| |
|![alt text](pictures/wf_simple.png) | |


Those issues of workflow-based execution can easily get even worse if considering that the components might have external data dependencies and be deployed on a distributed, heterogeneous infrastructure.


| *A distributed workflow example:*       | |
|-----------------------------------------| |
|![alt text](pictures/wf_distributed.png) | |


Dealing with such complex application workflows can be automated with the help of a workflow management system (**WMS**) that deals with the above-mentioned issues pertained to the workflow-based applications support. However, the wide-spread WMS normally rely on an excessive middleware stack which makes its integration with productional infrastructures (like HPC) very complicated, as it normally requires special permissions etc.


The **WMS-light** system follows a different approach - it aims to provide a flexible framework of light-weight and portable software tools and services that can simplify the development and execution of workflow-based applications while possibly introducing the lowest overhead to the inrastructure resources' system software stack. This approach is achieved by leveraging a high level of decentralization of the major workflow management components, their service-orientation as well as a distributed data management platform for tracking all application-specific data in a unified way.

The further content summarizes the major features, design and usage strategy of the **WMS-light** solution.


## 2 DESIGN AND BASIC FUNCTIONALITY

The **WMS-light** framework offers a "3-steps" strategy to get the workflow-based application execution on a distributed infrastructure completed, as illustrated below.


| *The workflow execution strategy:*      | |
|-----------------------------------------| |
|![alt text](pictures/3-steps.png)        | |


All user-defined models (component, infrastructure and deployment) rely on the transparable and user-friendly specification in JSON format. Specification details can be obtained in the "Full Documentation" section of this documentation.


The WMS-light design is depicted below.

| *WMS-light design:*                     | |
|-----------------------------------------| |
|![alt text](pictures/wms_design.png)     | |



The **WMS-light** framework includes the following components:

  * **Resource Manager Service** - is a service that is used to set up and control the infrastructure for the workflow components execution. The service provides front-ends (via a *RESTful API*), through which the users can manage (register, edit, etc.) the infrastructure resources, as well as back-ends (via *Node.js*) for the graphical representation of the infrastructure status.
  * **Application Manager Service** - is a service that is used to register new application workflows and their deployment configurations. The service provides front-ends (via a *RESTful API*), through which the users can submit new JSON specifications of workflows and deployment plans.
  * **Execution Manager Service** - is a service that is used to register and control new executions for the previously submitted workflows (through the Application Manager). The service provides front-ends (via a *RESTful API*), through which the users can submit new JSON specifications of execution scenarios (containing an application specification and the corresponding deployment plan), as well as back-ends (via *Node.js*) for the graphical representation of the entire application's and its components' status.
  * **Monitoring Client** - is a tiny software daemon that is running on the infrastructure resources and gathering the data that are related to performance of both hardware and software (workflow components) running on those resources
  * **Data Layer** - is a common data storage used by all other WMS-light services. It provides a front-end (via a *RESTful API*) for setting and querying any service-related data (in JSON format).


## 3 SOFTWARE DEPENDENCIES

The machine that shall hosts all the WMS-light services (the WMS server) shall have the following components installed:
  * Compilers: GNU g++, Java
  * Distributed data analytics engine: [Elasticsearch](https://www.elastic.co/)
  * Linux command-line tools wget, boost, boost-deve, curl, curl-devel
  * For the graphical representation of the major interfaces: nodejs, npm

For a quick start, we recommend to use the Docker environment (see the corresponding section below) that contains all major services and software environment for running them.

## 4 BASIC CONCEPTS

### 4.1 Application Execution Model

The typical **WMS-light** application consists of components (any user-defined binaries or scripts), which can be data-dependend on each other (i.e. according to a "producer-consumer" strategy). Those data dependencies are treated as a control flow that stears the execution of application components - every "consumer" component cannot be started before its "producers" (other components with outcoming data dependencies targeting the current one) have been completed and the data dependencies resolved. 

This approach can simplify the development of the application components by allowing them to assume that the data of all "producer"-components are locally available and can be red directly from the local file system. This approach is achieved by implementing the concept of *sandboxes* - special folders (located locally to the software components) containing the data obtained from the predecessing "producers" (the **input-sandbox**) as well as the data which the component wishes to be propagated to all its followers (i.e. "consumers") after its execution has been completed, to be placed into the **output-sandbox**. Both input- and output-sandboxes are created by the workflow execution framework, independently for every application run (i.e. for a specific UEID).

This idea is illustrated in the figure below.

| *Realization of data dependencies between workflow components:*     | |
|---------------------------------------------------------------------| |
|![alt text](pictures/execution_model.png)                            | |


### 4.2 Resource Manager (RM)

RM is a service that is responsible for setting up and controlling the infrastructure for the application execution. 

The RM front-ends can be accessed via the following RESTful API:


```bash
  .../rm - the index for infrastructure specification by the users

  .../rm/_doc/{$ID} - details of the particular infrastructure resource

  .../rm_stat - the index for monitoring information of infrastructure resources

  .../rm_stat/_doc/{$ID} - monitoring details for the particular infrastructure resource
```

Here is {$ID} - the ID of the resource (e.g. "host_1")

The infrastructure specification has the following format:

```bash
    cat local_host.json 
    {
        "id":"host_1",
        "address":"127.0.0.1",
        "cpus":"1",
        "cores/cpu":"8",
    	"desc":"laptop"
    }
```

### 4.2 Application Manager (AM)

AM is a service that is responsible for the registration of the application (workflow) and deployment configurations, to be submitted by the users to the AM's front-end.

The AM's front-end exposes the following API for registering application specification:
```bash
  .../am - the index for workflow specifications by the users

  .../am/_doc/{$ID} - details of the particular workflow configuration
```

Here is {$ID} - the ID of the workflow (e.g. "My_workflow_1").

The workflow specification has the following format:

```bash
    cat simple_workflow.json 
    {
        "id":"Application_1",
        "descr":"My test application",
        "components":
        [
    	{
    	    "id":"component_1",
    	    "descr":"pre-procesing",
    	    "type":"serial",
    	    "exec":"preprocessor.sh",
    	    "depend":"none"
    	},
    	{
    	    "id":"component_2",
    	    "descr":"simulator",
    	    "type":"mpi",
    	    "nr_mpi":"5",
    	    "exec":"simulator.sh",
    	    "depend":"component_1"
    	},
    	{
    	    "id":"component_3",
    	    "descr":"post-procesing",
    	    "type":"serial",
    	    "exec":"postprocessor.sh",
    	    "depend":"component_2"
    	}
        ]
    }
```

The AM's front-end also exposes the following API for registering deployment specification:

```bash
  .../am_depl - the index for deployment specifications by the users

  .../am_depl/_doc/{$ID} - details of the particular deployment configuration
```

Here is {$ID} - the ID of the deployment configuration (e.g. "Local_deployment_1").

The deployment specification has the following format:

```bash
   cat simple_deployment.json 
    {
        "id":"Deployment_1",
        "descr":"Standart Deployment",
        "components":
        [
    	{
    	    "id":"component_1",
    	    "host":"host_1",
    	    "dir":"/tmp",
    	    "auth":"cert"
    	},
    	{
    	    "id":"component_2",
    	    "host":"host_2",
    	    "dir":"/tmp",
    	    "auth":"cert"
    	},
    	{
    	    "id":"component_3",
    	    "host":"/tmp",
    	    "dir":"$HOME/application3",
    	    "auth":"cert"
    	}
        ]
    }
```


### 4.3 Execution Manager (EM)

EM is a service that is responsible for the execution of the application on the infrastructure with a user-defined deployment configuration. The infrastructure, application (workflow) and deployment configurations are fed from the RM and AM services (see above).

EM exposes the following interfaces:

```bash
  .../em - the index for static executions specifications

  .../em/_doc/{$ID} - details of the particular execution configuration (contains wf_id, depl_id)

  .../em/_doc/executions - details of all previously submitted  execution configurations

  .../em_stat - the index for monitoring information about the executions

  .../em_stat/_doc/{$ID} - monitoring details of the particular execution
```
Here is {$ID} - the ID of the execution (e.g. "2020-03-31T11:30:00.000") 

The execution request contains IDs of the applcation workflow and deployment plan (as assigned by the AM).



## 5 Quick-Start Guide (with Docker)

The easiest way to get started with using WMS-light is to use a Docker container, which contains the major WMS-light components with all their software dependencies.
In order to start WMS-light within the Docker container:

  * install Docker on the machine where WMS-light shall be running.
  * build a Docker container from the provided Docker file (see in the folder "Docker" of the WMS-lite distribution). You might also want to use the provided script ./build_image.sh.
  * run the built Docker container, e.g. with the help of the script ./run_container.sh.
  * start a new session of the Docker container environment, using the command sudo docker exec -it <container_id> bash.
  * change to the demo directory "cd WMS-light/Demo"
  * run demo "./run_demo.sh"
  * optionally: discover details on the executions from the web-backend of the Execution Manager. For this, open a new browser window and point it to http://localhost:8700/em_list.html)
  
| *Accessing workflow information from the WMS-Engine back-end:*  | |
|-----------------------------------------------------------------| |
|![alt text](pictures/nodejs_backend.png)                         | |


## 6 Developer Guide

*Note:* The WMS-light path is specified as *$WMS_HOME* below.

### 6.1 Deployment of the major WMS-light services
#### Setup of the Data Layer on the WMS-Server

On the WMS-Engine host machine (further referred as WMS-Server), setup Elasticsearch and point the environmental variable *$ES_HOME* to the main folder of your Elasticsearch installation, e.g.:
```bash
    export ES_HOME=$HOME/Tools/Elasticsearch/elasticsearch-7.6.1
```
Run the Elasticsearch server, e.g.:
```bash
    $ES_HOME/bin/elasticsearch
```
In a new terminal session, test the Elasticsearch server (install the *curl* utility if not yet done), e.g.:
```bash
    curl -X GET http://localhost:9200/
```
Perform initial setup of the Data Layer functionality
```bash
    $WMS_HOME/Data_Layer/setup.sh
```

#### Setup of the Monitoring Client on the Infrastructure-Resources

The workflow management requires that on every infrastructure resource that hosts a software component of the workflow an instance of the Monitoring Client shall be deployed.

As a prerequisite action, ensure that a [CURL library](https://curl.haxx.se/download.html) has been downloaded and installed:
```bash
    ./configure --prefix=${CURL_HOME}; 
    make; 
    make install; 
    make install all
```
Point the *$CURL_HOME* environmental variable to the head folder of the CURL installation. Execute the *setup.sh* script. Installation is completed.

Run the Client as follows:
```bash
    $WMS_HOME/Monitoring_Client/monitoring
```
Perform the next steps in a new terminal session.

### 6.2 Registration of a new Infrastructure Resource by the Resource Manager's Frontend

The workflow execution infrastructure is considered as a set of distributed resources, each of those has to be registrered by the WMS-light's Resource Manager. For test purposes, try one of dummy resources provided in template in $WMS_HOME/Resource_Manager/Resource_Examples, e.g.:
```bash
    $WMS_HOME/Resource_Manager/Frontend/add_resource.sh $WMS_HOME/Resource_Manager/Frontend/Resource_Examples/local_host.json
```

### 6.3 Registration of a new Application Workflow by the Application Manager's Frontend

Every new user's worflow has to be registered (only once) at the Application Mananager prior to its deployment on the infrastructure.

For test purposes, try one of dummy workflow specifications provided in template in $WMS_HOME/Application_Manager/Workflow_Examples, e.g.:
```bash
    $WMS_HOME/Application_Manager/Frontend/add_workflow.sh $WMS_HOME/Application_Manager/Frontend/Workflow_Examples/simple_workflow.json
```

### 6.4 Registration of a new Deployment Configuration by the Application Manager's Frontend

A deployment configuration (for one of the previously submitted workflow scenarios) have to be defined prior to the execution and registered at the Application Manager as follows:

```bash
    $WMS_HOME/Application_Manager/Frontend/add_deployment.sh $WMS_HOME/Application_Manager/Frontend/Deployment_Examples/simple_deployment.json
```

### 6.5 Submission of Deployment Request to the WMS-Engine



<!-- Next are the setup of the Monitoring Client-->
### 6.6 Monitoring Client

Run the following command for help:
```bash
    ./monitoring -h
```


to run the monitoring on the same server node and port keeps as the default port, just run the command:

```bash
./monitoring ;
```

in other more general case, to run the monitoring just use the next command,
where the parameter 's' provides the server address where it is running the elasticsearch,
and the parameter 'p' provides the port, which is only needed to provide when be differnt than the default value of 9200.

```bash
./monitoring -s xx.xx.xx.xx. -p 9200;
```


### 6.7 WMS Engine

```bash
cd Engine;
bash run_demo.sh;
```

## Main Contributors

**Montanana, Jose Miguel, HLRS**
+ [github/jmmontanana](https://github.com/jmmontanana)

**Cheptsov, Alexey, HLRS**
+ [github/alexey-cheptsov](https://github.com/alexey-cheptsov)

## License
Copyright (C) 2020 University of Stuttgart

[Apache License v2](LICENSE).
